* OMEGA for SAS Version 1.0;
* current as of 14 April 2021;

* Copyright 2019 by Andrew F. Hayes;
* OMEGA for SAS requires PROC IML to function;

* Distribution of this code in any form, except through afhayes.com is prohibited;
* without the written permission of the copyright holder;

* THIS SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND;
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF;
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT;
* IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, ;
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT;
* OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE ;
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE ;
* USE OF THIS SOFTWARE IMPLIES AGREEMENT WITH THESE TERMS ;

*computes covariance matrix and extracts diagonals as var and sd;
%macro makcov (dat=);
	cov = (datac`*(I(n)-(1/n)*constant*constant`)*datac)*(1/(n-1));
	var = vecdiag(cov);
	sd = sqrt(var); 
%mend makcov;

*constructs a bootstrap sample;
%macro boot3 (dat=);
	v=floor((ranuni(j(n,1,&seed)))*n)+1;
	bootsamp=&dat[v,];
%mend boot3;
*implements the efa-ml approach and prints the results;
%macro omega2 (items=,stand=0);	

	ods select FactorPattern;
	ods output FactorPattern=sloads;
	proc factor data=&data
  		nfact=1
  		method=ml;
	run;
	proc iml;
		use sloads;read all var _all_ into data;
		use &data;read all var{&items} into rawd;vnames={&items};
		slambda=data;n = nrow(rawd);datac=slambda;constant = j(n,1,1);
		vars = vecdiag((rawd`*(I(n)-(1/n)*constant*constant`)*rawd)*(1/(n-1)));
		means = (rawd[+,]/n)`;
		sums = rawd[,+];var=(sums`*(I(n)-(1/n)*constant*constant`)*sums)*(1/(n-1));
		lambda=slambda#sqrt(vars);
		err=(1-(slambda#slambda))#vars;
		reliab = (lambda[+,]*lambda[+,])/(lambda[+,]*lambda[+,]+err[+,]);

		print "********************** OMEGA Procedure for SAS Beta Version 0.1 ********************";
		print "                             Written by Andrew F. Hayes                            ";
		print "This estimate of omega is based on the factor loadings of a forced single-factor maximum-likelihood factor analysis usings SAS's built-in FACTOR procedure.";
		print reliab [label="Reliability:" colname = "Omega" format=8.3];
		lambda = means || sqrt(vars) || lambda || err;
		clabrynth = "Mean" || "SD" || "Loading" || "Error Var";
		print lambda [label = "Item means, standard deviations, and estimated loadings" colname=clabrynth rowname=vnames format=8.3];
		err=(1-(slambda#slambda));
		if (&stand=1)then;do;
			sres = slambda||err;
			reliab=(slambda[+,]*slambda[+,])/(slambda[+,]*slambda[+,]+err[+,]);
			print "--------------";
			print "Standardized results:";
			print reliab [label="Reliability:" colname="Omega" format=8.3];
			clab="Loading"||"ErrorVar";
			print sres [label="Loadings and error variances" colname=clab rowname=vnames format=8.3];
		end;
	quit;
%mend omega2;

%macro omegac (dat=);
	totalsum=&dat[,+];nitems=ncol(&dat);datac=&dat||totalsum||totsumf;
	cnt=1;
	if (bootstrt = 0) then;do;
		if (alpha=0) then;do;
			items3=j((nitems*(nitems-1)*(nitems-2)/6),nitems,0);
			items3r=j(nrow(items3),3,0);
			do ic=1 to (nitems-2);
				do jc=(ic+1) to (nitems-1);
					do kc=(jc+1) to nitems;
						items3[cnt,ic]=1;
						items3[cnt,jc]=1;
						items3[cnt,kc]=1;
						items3r[cnt,1]=ic;
						items3r[cnt,2]=jc;
						items3r[cnt,3]=kc;
						cnt=cnt+1;
					end;
				end;
			end;
		end;
	end;
  %makcov (dat=datac);
	rminus=cov[nrow(cov),(nrow(cov)-1)]/(sd[nrow(cov)]*sd[nrow(cov)-1]);
	sumvar=var[(nitems+1),1];
	alph=(nitems/(nitems-1))*((var[(nitems+1),1]-sum(var[1:nitems,1]))/sumvar);
	
	if (alpha=0) then;do;
		num=j((nitems*(nitems-1)*(nitems-2)/6),nitems,0);den=num;
		do ic=1 to nrow(items3);
			num[ic,items3r[ic,1]]=cov[items3r[ic,1],items3r[ic,2]]*cov[items3r[ic,1],items3r[ic,3]];
			num[ic,items3r[ic,2]]=cov[items3r[ic,2],items3r[ic,1]]*cov[items3r[ic,2],items3r[ic,3]];
			num[ic,items3r[ic,3]]=cov[items3r[ic,3],items3r[ic,2]]*cov[items3r[ic,3],items3r[ic,1]];
			den[ic,items3r[ic,1]]=cov[items3r[ic,2],items3r[ic,3]];
			den[ic,items3r[ic,2]]=cov[items3r[ic,1],items3r[ic,3]];
			den[ic,items3r[ic,3]]=cov[items3r[ic,2],items3r[ic,1]];
		end;
		chk=num[+,]/den[+,];
		negchk=sum(chk<0);
		
		if (negchk>0) then;do;
			omeg=-99;lambda=(j(nitems,1,-99))`;negprob=negprob+1;

			if (first=1) then;do;
				m999=1;
			end;

			if (first=2) then;do;
				gdboot=gdboot-1;
				bdboot=bdboot+1;
			end;
		end;

		if (negchk=0) then;do;
			lambda=sqrt(chk);
			omeg=(lambda[,+]*lambda[,+])/var[(nrow(var)-1),1];

			if (omeg > 1) then;do;
				omeg=-99;gtoprob=gtoprob+1;

				if (first=2) then;do;
					gdboot=gdboot-1;
					bdboot=bdboot+1;
				end;

				if (first=1) then;do;
					m999=1;
				end;
			end;
		end;
	end;
	model="Omega";
	if (alpha=1) then;do;
		omeg=alph;
		model="alpha";
	end;
%mend omegac;
*here is the main macro;
%macro omega (data=,items=,alpha=0,sort=0,boot=0,seed=0,subsets=0,covmat=0,save=0,file=savefile,ha=0,loo=0,stand=0);
	options pagesize=32767;
	%if ((&ha = 0) & (&alpha=0)) %then %do;
	  %omega2 (items=&items, stand=&stand); 
	%end;
	%if ((&ha=1) | (&alpha=1) | (&covmat=1)) %then %do;
		proc iml;*startprociml;
			*setting up base values;
			ha=floor(&ha);stand=floor(&stand);
			use &data;read all var{&items} into data;vnames={&items};*read in data;
			boots=floor(&boot);subsets=(&subsets=1);
			if ((&boot < 1000) & (&boot ^= 0)) then;do;*check boot value;
				boots=1000;
			end; 
			toofew=0;toomany=0;sub999=0;mxboots=2*boots;bootstrt=0;bootok=0;savedit=0;covmat=floor(&covmat);
			save=(&save=1);items=ncol(data);loo=(&loo=1);alpha=(&alpha=1);sortc=2-(floor(&sort));*t;
	    	m999=0;first=1;negprob=0;gtoprob=0;criterr=0;

			*set seed;
			if (&seed ^= 0) then;do;
				sn=&seed;
	    	end;

			*check for proper number of items;
			if (((alpha = 1) & (items < 2)) | ((alpha = 0) & (items < 3))) then;do;
	    		criterr=1;toofew=1;
	  		end;
			
			*check for missing data;
			do j = 1 to ncol(data);
	      		do i = 1 to nrow(data);if data[i,j]=. then;do;data[i,j]=99999;end;end;
	    	end;
			*purging missing data;
			j=1;do i = 1 to nrow(data);
	        	tmprow=data[i,];tmprow=(tmprow=99999);tmprow=tmprow[,+];
		    	if (tmprow=0) then;do;data[j,]=data[i,];j=j+1;end;
	    	end;
			data=data[1:(j-1),];n=nrow(data);constant=j(n,1,1);totsumf=data[,+];means=(data[+,]/n)`;

			*start computations;
			if (criterr=0) then;do;
			  %omegac (dat=data);
				corr=vecdiag(1/sqrt(var))*cov*vecdiag(1/sqrt(var))`;
				corr=corr[1:(nrow(corr)-2),1:(nrow(corr)-2)];
				sumvfull=var[nitems+1,1];summn=totalsum[+,]/nrow(totalsum);
				sdd=sd;corr2=corr;
				do i=1 to nitems;
					do j=1 to nitems;
						corr[i,j]=cov[i,j];
					end;
				end;

				first=0;omegobs=omeg;
				
				if (omegobs=-99) then;do;
					boots=0;
				end;

				model2=model;

				if (alpha=0) then;do;
					lambobs=lambda;
				end;
				gdboot=0;bdboot=0;complete=0;

				if (boots>0) then;do;
					bootd=j(mxboots,1,-99);
					negprob=0;gtoprob=0;first=2;
					do i=1 to mxboots UNTIL (gdboot=boots);
						gdboot=gdboot+1;
						%boot3 (dat=data);
						%omegac (dat=bootsamp);
						bootstrt=1;bootd[i,1]=omeg;
					end;	

					if (gdboot=boots) then;do;
						complete=1;bootok=1;
						tmp=bootd;tmp[rank(bootd),]=bootd;bootd=tmp;
						bootd=bootd[(mxboots-gdboot+1):nrow(bootd),1];
						llci=bootd[floor(.025*nrow(bootd)),1];
						ulci=bootd[(floor(.975*nrow(bootd))+1),1];
						seboot=bootd-(bootd[+,]/nrow(bootd));
						seboot=sqrt(((seboot`*seboot)[+,])/(nrow(bootd)-1));
						omegobs=omegobs||seboot||llci||ulci;
						model2=model||"BootSE"||"BootLL95"||"BootUL95";
						bootok=1;
					end;
				end;
				bootstrt=0;
				
				if (ha=1 | alpha=1) then;do;
	   		    	print "************************ OMEGA Procedure for SAS Version 1.0 **********************";
					print "                             Written by Andrew F. Hayes                            ";
	  				if (alpha=0) then;do;
						print "This estimate of omega is based on the approximate and closed-form solution to the computation 
						*of loadings described in Hancock, G.R., and An, J. (2020).";
						print "A closed-form alternative for estimating omega reliability under unidimensionality. Measurement: 
						*Interdisciplinary Research and Perspectives.";
					end;

					print omegobs [label = "Reliability" colname=model2 format=8.3];
					clbs="Mean"||"SD";
					descr=means||sdd[1:nitems,1];

					if (alpha ^= 1) then;do;
						descr=descr||(lambobs`);
						clbs=clbs||"Loading";
					end;

					if ((items > 3 & alpha = 0) | (items > 2 & alpha=1)) then;do;
						first=1;datat=j(n,(items-1),0);omeglo=j(nitems,1,0);

						do ilo=1 to items;
							cnt=1;
							do jlo=1 to items;
								if (jlo ^= ilo) then;do;
									datat[,cnt]=data[,jlo];cnt=cnt+1;
								end;
							end;
							%omegac (dat=datat);
							omeglo[ilo,1]=omeg;
						end;

						if (loo=1) then;do;
							descr=descr||omeglo;
						end;

						clbs=clbs||model;

						if (alpha=0) then;do;
							if (loo=1) then;do;
								print descr [label="Item means, standard deviations, estimated loadings, and reliability if item deleted:" rowname=vnames colname=clbs format=8.3];
							end;
							if (loo=0) then;do;
								print descr [label="Item means, standard deviations, and estimated loadings:" rowname=vnames colname=clbs format=8.3];
							end;
						end;

						if (alpha=1) then;do;
							if (stand=1) then;do;
								meanr=((sum(corr2[,+])-items)/2)/((items*(items-1))/2);
								salpha=(items*meanr)/(1+(items-1)*meanr);
								print salpha [label="Standardized alpha:" format=8.3];
							end;

							if (loo=1) then;do;
								print descr [label="Item means, standard deviations, and reliability if item deleted:" rowname=vnames colname=clbs format=8.3];
							end;

							if (loo=0) then;do;
								print descr [label="Item means and standard deviations:" rowname=vnames colname=clbs format=8.3];
							end;
						end;
					end;
				end;
				
				if (covmat=1) then;do;
					print corr [label="Item variances (diagonal) and inter-item correlations (lower), and covariances (upper)" rowname=vnames colname=vnames format=8.3];
				end;

				*start all subsets;
				if (subsets=1 & ((ha=1) | (alpha=1))) then;do;
					if (((items>3) & (alpha=0)) | (items>2 & alpha=1)) then;do;
						first=1;x=j(1,items,0);k=0;y=j(((2##items)-1)-items,items,0);

						do j=1 to ((2##items)-1);
							x[1,1]=x[1,1]+1;
							do i=1 to items;
								if (x[1,i]=2) then;do;
									x[1,i]=0;
									x[1,i+1]=x[1,i+1]+1;
								end;
							end;

							if (x[,+] > (2-alpha)) then;do;
								k=k+1;y[k,]=x;
							end;
						end;

						y=y[1:(k-0),];subnitem=y[,+];relsub=j(nrow(y),2,0);

						do ilo=1 to nrow(y);
							datat=j(n,subnitem[ilo,1],0);
							cnt=1;
							do jlo=1 to items;
								if (y[ilo,jlo]=1) then;do;
									datat[,cnt]=data[,jlo];cnt=cnt+1;
								end;
							end;
						  %omegac (dat=datat);
							relsub[ilo,1]=omeg;relsub[ilo,2]=rminus;
						end;
						y=y||relsub||subnitem;
						temp=y;
						temp[(nrow(y)+1-rank(y[,(ncol(y)-sortc)])),]=y;
						y=temp;
						y=y[,1:(ncol(y)-1)];
						vnames2=vnames||model||"r";

						if (alpha=0) then;do;
							print y [label="Reliability and subscale-full scale correlations (r) for all possible subscales with 3 or more items (1=item included, 0=not included):" colname=vnames2 format=7.3];
						end;

						if (alpha=1) then;do;
							print y [label="Reliability and subscale-full scale correlations (r) for all possible subscales with 2 or more items (1=item included, 0=not included):" colname=vnames2 format=7.3];
						end;

						res=j(items,3,0);
						sub999=((1-(y[,(items+1)]=-99))=0);sub999=sub999[+,];

						if (sub999=0) then;do;
							do i=1 to items;
								pre=j(nrow(y),1,1)||y[,i];
								b=inv(pre`*pre)*(pre`)*y[,(items+1)];
								res[i,]=b[1,1]||(b[1,1]+b[2,1])||b[2,1];
							end;
							examlab="Without"||"With"||"Gain";
							print res [label="Mean reliability of all possible subset scales with and without item:" colname=examlab rowname=vnames format=8.3];
						end;

						if (save=1) then;do;
							vnames=vnames||"reliab"||"r";
							create &file from y [colname=vnames];append from y;close &file;
							savedit=1;
						end;
					end;
				end;
			end;

			print "**************************** ANALYSIS NOTES AND ERRORS *****************************";
			if (toofew=1) then;do;
				print "ERROR: Too few items for the analysis you have requested.";
			end;
			*if (toomany=1) then;*do;
			*	print "ERROR: Too many items. 15 is the maximum allowed.";
			*end;
			if (criterr=0) then;do;
				print n [label = "Number of cases:"];
				if (m999=1) then;do;
					print "NOTE: -99 indicates that the algorithm for calculating this quantity generated an error. This is";
					print "probably due to including an item in the scale that is very weakly or negatively correlated with the others.";
				end;
				if (sub999 > 0) then;do;
					print "NOTE: Because reliability could not be estimated for some subset scales, the table of mean reliability"; 
					print "with and without item is not generated";
				end;
				if (bootok = 1) then;do;
					if ((boots >= 1000) & (&boot >= 1000)) then;do;
						print boots [label = "NOTE: The number of bootstrap samples for the 95% confidence interval:"];
					end;
					if (&boot < 1000) then;do;
						print "NOTE: 1000 bootstrap samples is the minimum, so 1000 was used for construction of the 95% confidence interval.";
					end;
					if (bdboot > 0) then;do;
						print "NOTE: Some bootstrap samples had to be replaced due to computational problems. Interpret confidence"; 
						print bdboot [label ="interval with caution. The number of bootstrap samples replaced was:"];
					end;
					if (&seed > 0) then;do;
						print sn [label = "Seed used for the random number generator:"];
					end;
				end;
				if ((complete=0) & (boots > 0)) then;do;
					print "NOTE: Bootstrapping could not be completed due to computational problems so no bootstrap confidence interval is provided.";
				end;
				if (savedit=1) then;do;
					print "NOTE: A file of all subsets was generated at your request";
				end;
			end;
		quit; *stop prociml;
	%end;
	proc delete data=sloads;run;
	options pagesize=54;
%mend omega;

