﻿* Encoding: UTF-8.

/* Version 1.0 */.
/* 14 April 2021 */.
/* Written by Andrew F. Hayes */.
/* www.afhayes.com */.

/* THIS SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND */.
/* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF */.
/* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT */.
/* IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, */.
/* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT */.
/* OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE */.
/* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE */.
/* USE OF THIS SOFTWARE IMPLIES AGREEMENT WITH THESE TERMS */.

/* this computes a covariance matrix as well as extracts the diagonals as var and sd */.
DEFINE makcov (dat=!charend('/')).
compute cov = (t(datac)*(ident(n)-(1/n)*constant*t(constant))*datac)*(1/(n-1)).
compute var=diag(cov).
compute sd=sqrt(var).
!ENDDEFINE.

/* this constructs a bootstrap sample */.
DEFINE boot3 (dat=!charend('/')).
compute v=trunc(uniform(n,1)*n)+1.
compute bootsamp=!dat(v,:).
!ENDDEFINE.

/* The omega2 macro implements the EFA-ML approach and prints the results */.
DEFINE omega2 (items=!charend('/')/stand=!charend('/') !default(0)/decimals=!charend('/') !default (F8.3)).
dataset name temp.
dataset copy omegcopy.
dataset activate omegcopy.
count miss=!items (missing).
select if (miss=0).
!LET !b=' '.
!LET !c=' '.
!DO !i !IN (!items)
!LET !b = !CONCAT(!b,'+',!i).
!DOEND
compute sumscore=sum(!b).
execute.
dataset declare omegare2.
 oms/select tables/destination format=sav outfile = omegare2 
 /if commands = ['Descriptives'] subtypes = ['Descriptive Statistics'].
descriptives variables = sumscore !items/statistics=variance mean.
omsend.
dataset close omegcopy.
dataset declare omegare1.
  oms/select tables/destination format=sav outfile = omegare1 
 /if commands = ['Factor Analysis'] subtypes = ['Factor Matrix'].
factor variables = !items/
 criteria factors(1) iterate(100)/extraction=ml/method=correlation.
omsend. 
dataset activate omegare2.
!DO !i !IN (!items)
compute !i = 999.
!DOEND
execute.
match files/file=*/drop Command_ Subtype_ Label_ Var1 N.
dataset activate omegare1.
!DO !i !IN (!items)
compute !i = 999.
!DOEND
compute mean=999.
add files/file=*/rename (@1 = data)/file=omegare2/Rename (variance=data).
match files/file=*/drop Command_ Subtype_ Label_ Var1.
execute.
matrix.
get data/file=*/names=vnames/missing=omit.
compute vnames=vnames(:,2:(ncol(vnames)-1)).
compute slambda=data(1:((nrow(data)-1)/2),1).
compute vars=data((((nrow(data)-1)/2)+2):nrow(data),1).
compute means=data((((nrow(data)-1)/2)+2):nrow(data),ncol(data)).
compute lambda=slambda&*sqrt(vars).
compute var=data((((nrow(data)-1)/2)+1),1).
compute err=(1-(slambda&*slambda))&*vars.
/* compute reliab=(csum(lambda)*csum(lambda))/var */.
compute reliab=(csum(lambda)*csum(lambda))
/(csum(lambda)*csum(lambda)+csum(err)).
print/title="************************* Written by Andrew F. Hayes **************************".
print/title="*************************        afhayes.com         **************************"/space=0.
print/title="This estimate of omega is based on the factor loadings of a forced single-factor".
print/title="maximum likelihood factor analysis using SPSSs built in FACTOR procedure."/space=0.
print reliab/title="Reliability:"/clabels="Omega"/format=!decimals.
compute lambda={means,sqrt(vars),lambda, err}.
print lambda/title="Item means, standard deviations, and estimated loadings:"/clabels="Mean", "SD", "Loading", "ErrorVar"/rnames=vnames/format=!decimals.
compute err=(1-(slambda&*slambda)).
do if (!stand=1).
  compute sres={slambda,err}.
  compute reliab=(csum(slambda)*csum(slambda))/(csum(slambda)*csum(slambda)+csum(err)).
  print/title="--------------".
  print/title="Standardized results:"/space=0.
  print reliab/title="Reliability:"/clabels="Omega"/format=!decimals.
  print sres/title="Loadings and error variances"/clabels="Loading", "ErrorVar"/rnames=vnames/format=!decimals.
end if.
end matrix.
dataset activate temp.
dataset close omegare1.
dataset close omegare2.
!ENDDEFINE.

/* The omegac macro implements the computation of alpha or omega using the HA algorithm */.
DEFINE omegac (dat=!charend('/')).
compute totalsum=rsum(!dat).
compute nitems=ncol(!dat).
compute datac={!dat,totalsum, totsumf}.
compute cnt=1.
do if (bootstrt=0).
  do if (alpha=0).
  compute items3=make((nitems*(nitems-1)*(nitems-2)/6),nitems,0).
  compute items3r=make(nrow(items3),3,0).
  loop ic=1 to (nitems-2).
    loop jc=(ic+1) to (nitems-1).
      loop kc = (jc+1) to nitems.
        compute items3(cnt,ic)=1.
        compute items3(cnt,jc)=1.
        compute items3(cnt,kc)=1.
        compute items3r(cnt,1)=ic.
        compute items3r(cnt,2)=jc.
        compute items3r(cnt,3)=kc.
        compute cnt=cnt+1.
      end loop.
    end loop.
  end loop.
  end if.
end if.
makcov dat=datac.
compute rminus=cov(nrow(cov),(nrow(cov)-1))/(sd(nrow(cov))*sd(nrow(cov)-1)).
compute sumvar=var((nitems+1),1).
compute alph=(nitems/(nitems-1))*((var((nitems+1),1)-csum(var(1:nitems,1))) /sumvar).
do if (alpha = 0).
  compute num=make((nitems*(nitems-1)*(nitems-2)/6),nitems,0).
  compute den=num.
  loop ic = 1 to nrow(items3).
    compute num(ic,items3r(ic,1))=cov(items3r(ic,1),items3r(ic,2))*cov(items3r(ic,1),items3r(ic,3)).
    compute num(ic,items3r(ic,2))=cov(items3r(ic,2),items3r(ic,1))*cov(items3r(ic,2),items3r(ic,3)).
    compute num(ic,items3r(ic,3))=cov(items3r(ic,3),items3r(ic,2))*cov(items3r(ic,3),items3r(ic,1)).
    compute den(ic,items3r(ic,1))=cov(items3r(ic,2),items3r(ic,3)).
    compute den(ic,items3r(ic,2))=cov(items3r(ic,1),items3r(ic,3)).
    compute den(ic,items3r(ic,3))=cov(items3r(ic,2),items3r(ic,1)).
  end loop.
  compute chk=csum(num)&/csum(den).
   compute negchk=rsum(chk<0).
   do if (negchk>0).
     compute omeg=-99.
     compute lambda=t(make(nitems,1,-99)).
     compute negprob=negprob+1.
     do if (first=1).
       compute m999=1.
     end if.
     do if (first=2).
       compute gdboot=gdboot-1.
       compute bdboot=bdboot+1.
     end if.
   end if. 
   do if (negchk=0).
    compute lambda=sqrt(chk).
    compute omeg=(rsum(lambda)*rsum(lambda))/var((nrow(var)-1),1).
    do if (omeg > 1).
      compute omeg=-99.
      compute gtoprob=gtoprob+1.
      do if (first=2).
        compute gdboot=gdboot-1.
        compute bdboot=bdboot+1.
      end if.
      do if (first=1).
       compute m999=1.
     end if.
    end if.
  end if.
end if.
compute model="Omega".
do if (alpha=1).
  compute omeg=alph.
  compute model="alpha".
end if.
!ENDDEFINE.

/* here is the main macro */.
SET MXLOOP=1000000000.
DEFINE omega (items=!charend('/')/alpha=!charend('/') !default (0)/sort=!charend('/') !default(0)
  /boot=!charend('/') !default(0)/seed=!charend('/') !default(random)/subsets=!charend('/') !default(0)/
  covmat=!charend('/') !default(0)/save=!charend('/') !default(0)/ha=!charend('/') !default(0)
  /loo=!charend('/') !default(0)/stand=!charend('/') !default(0)/decimals=!charend('/') !default(F7.3)/
   conf=!charend('/') !default(95)).
SET SEED=!seed.
!if (!ha=0 !and !alpha=0) !then
  preserve.
  set olang=english.
  omega2 items=!items/stand=!stand/decimals=!decimals.
  restore.
!ifend
!if (!ha = 1 !or !alpha=1 !or !covmat=1) !then
MATRIX.
compute ha=trunc(!ha).
compute stand=trunc(!stand).
get data/file=*/variables=!items/names=vnames/missing=omit.
compute boots=trunc(!boot).
compute subsets=(!subsets=1).
do if (!boot < 1000 and !boot <> 0).
  compute boots=1000.
end if.
compute toofew=0.
compute toomany=0.
compute sub999=0.
compute mxboots=2*boots.
compute n=nrow(data).
compute bootstrt=0.
compute bootok=0.
compute savedit=0.
compute covmat=trunc(!covmat).
compute save=(!save=1).
compute constant=make(n,1,1).
compute items=ncol(data).
compute loo=(!loo=1).
compute alpha=(!alpha=1).
compute sortc=2-(trunc(!sort)).
compute means=t(csum(data)/n).
compute totsumf=rsum(data).
compute m999=0.
compute first=1.
compute negprob=0.
compute gtoprob=0.
compute criterr=0.
compute conferr=0.
compute conf=trunc(!conf).
compute conf2=(1-(!conf/100))/2.
do if (conf <> 95 and conf <> 90 and conf <> 99).
  compute conf=95.
  compute conf2=.025.
  compute conferr=1.
end if.
do if ((alpha=1 and items < 2) or (alpha=0 and items < 3)).
  compute criterr=1.
  compute toofew=1.
end if.
do if (criterr=0).
  omegac dat=data.
  compute corr=mdiag(1/sqrt(var))*cov*t(mdiag(1/sqrt(var))).
  compute corr=corr(1:(nrow(corr)-2),1:(nrow(corr)-2)).
  compute sumvfull=var(nitems+1,1).
  compute summn=csum(totalsum)/nrow(totalsum).
  compute sdd=sd.
  compute corr2=corr.
  loop i = 1 to nitems.
    loop j = i to nitems.
      compute corr(i,j)=cov(i,j).
    end loop.
  end loop.
  compute first=0.
  compute omegobs=omeg.
  do if (omegobs=-99).
    compute boots=0.
  end if.
  compute model2=model.
  do if (alpha = 0).
    compute lambobs=lambda.
  end if.
  compute gdboot=0.
  compute bdboot=0.
  compute complete=0.
  do if (boots > 0).
    compute bootd=make(mxboots,1,-99).
    compute negprob=0.
    compute gtoprob=0.
    compute first=2.
    loop i = 1 to mxboots.
      compute gdboot=gdboot+1.
      boot3 dat=data.
      omegac dat=bootsamp.
      compute bootstrt=1.
      compute bootd(i,1)=omeg.
    end loop if (gdboot=boots).
    do if (gdboot=boots).
      compute complete=1.
      compute bootok=1.
      compute tmp = bootd.
      compute tmp(GRADE(bootd),:) = bootd.
      compute bootd= tmp.
      compute bootd=bootd((mxboots-gdboot+1):nrow(bootd),1).
      compute llci=bootd(trunc(conf2*nrow(bootd)),1).
      compute ulci=bootd((trunc((1-conf2)*nrow(bootd))+1),1).
      compute seboot=bootd-(csum(bootd)/nrow(bootd)).
      compute seboot=sqrt(csum(t(seboot)*seboot)/(nrow(bootd)-1)).
      compute omegobs={omegobs,seboot,llci,ulci}.
      compute model2={model,"BootSE", "BootLLCI", "BootULCI"}.
      compute bootok=1.
    end if.
  end if.
  compute bootstrt=0.
  do if (ha = 1 or alpha=1).
    print/title="************************* Written by Andrew F. Hayes **************************".
    print/title="*************************        afhayes.com         **************************"/space=0.
    do if (alpha=0).
      print/title="This estimate of omega is based on the approximate and closed-form solution".
      print/title="to the computation of loadings described in Hancock, G. R., and An, J. (2020)."/space=0.
      print/title="A closed-form alternative for estimating omega reliability under unidimensionality."/space=0.
      print/title="Measurement: Interdisciplinary Research and Perspectives."/space=0. 
    end if.
    print omegobs/title="Reliability:"/cnames=model2/format=!decimals.
    compute clbs={"Mean", "SD"}.
    compute descr={means,sdd(1:nitems,1)}.
    do if (alpha <> 1).
      compute descr={descr,t(lambobs)}.
      compute clbs={clbs,"Loading"}.
    end if.
    do if ((items > 3 and alpha=0) or (items > 2 and alpha=1)).
      compute first=1.
      compute datat=make(n,(items-1),0). 
      compute omeglo=make(nitems,1,0).
      loop ilo = 1 to items.
        compute cnt=1.
        loop jlo = 1 to items.
          do if (jlo <> ilo).
            compute datat(:,cnt)=data(:,jlo).
            compute cnt=cnt+1.
          end if.
        end loop.
        omegac dat=datat.
        compute omeglo(ilo,1)=omeg.
      end loop.
      do if (loo=1).
        compute descr={descr,omeglo}.
      end if.
      compute clbs={clbs,model}.
      do if (alpha=0).
        do if (loo=1).
          print descr/title="Item means, standard deviations, estimated loadings, and reliability if item deleted:"/rnames=vnames/cnames=clbs/format=!decimals.
        end if.
        do if (loo=0).
          print descr/title="Item means, standard deviations, and estimated loadings:"/rnames=vnames/cnames=clbs/format=!decimals.
        end if.
      end if.
      do if (alpha=1).
        do if (stand=1).
          compute meanr=((csum(rsum(corr2))-items)/2)/((items*(items-1))/2).
          compute salpha=(items*meanr)/(1+(items-1)*meanr).
          print salpha/title="Standardized alpha:"/format=!decimals.
        end if.
        do if (loo=1).
          print descr/title="Item means, standard deviations, and reliability if item deleted:"/rnames=vnames/cnames=clbs/format=!decimals.
        end if.
        do if (loo=0).
          print descr/title="Item means and standard deviations:"/rnames=vnames/cnames=clbs/format=!decimals.      
        end if.
      end if.
      /* compute sumdesc={summn,sqrt(sumvfull),sumvfull} */.
      /* print sumdesc/title="Sumscore Descriptives"/clabels = "Mean", "SD", "Var"/format=!decimals */.
    end if.
  end if.
  do if (covmat=1).
    print corr/title="Item variances (diagonal) and inter-item correlations (lower), and covariances (upper)"/rnames=vnames/cnames=vnames/format=!decimals.
  end if.
  /* start all subsets */.
  do if (subsets=1 and (ha=1 or alpha=1)).
    do if ((items > 3 and alpha=0) or (items > 2 and alpha=1)).
      compute first=1.
      compute x=make(1,items,0).
      compute k= 0.
      compute y=make(((2&**items)-1)-items,items,0).
      loop j = 1 to ((2&**items)-1).
        compute x(1,1)=x(1,1)+1.
        loop i = 1 to items.
          do if x(1,i) = 2.
            compute x(1,i)=0.
            compute x(1,i+1)=x(1,i+1)+1.
          end if.
        end loop.
        do if (rsum(x) > (2-alpha)).
          compute k = k + 1.
          compute y(k,:)= {x}.
        end if.
      end loop.
      compute y=y(1:(k-0),:).
      compute subnitem=rsum(y).
      compute relsub=make(nrow(y),2,0).
      loop ilo = 1 to nrow(y).
        compute datat=make(n,subnitem(ilo,1),0).
        compute cnt=1.
        loop jlo = 1 to items.
          do if (y(ilo,jlo)=1).
            compute datat(:,cnt)=data(:,jlo).
            compute cnt=cnt+1.
          end if.
        end loop.
        omegac dat=datat.
        compute relsub(ilo,1)=omeg.
        compute relsub(ilo,2)=rminus.
      end loop.
      compute y={y,relsub,subnitem}.
      compute temp = y.
      compute temp(   (nrow(y)+1-GRADE(y(:,(ncol(y)-sortc))))  ,:) = y.
      compute y = temp.
      compute y=y(:,1:(ncol(y)-1)).
      compute vnames2={vnames,model,"r"}.
      do if (alpha=0).
        print/title="Reliability and subscale-full scale correlations (r) for all possible".
        print y/title="subscales with 3 or more items (1=item included, 0=not included):"/cnames=vnames2/format=!decimals/space=0.
      end if.
      do if (alpha=1).
        print/title="Reliability and subscale-full scale correlations (r) for all possible".
        print y/title="subscales with 2 or more items (1=item included, 0=not included):"/cnames=vnames2/format=!decimals/space=0.
      end if.  
      compute res = make(items,3,0).
      compute sub999=((1-(y(:,(items+1))=-99))=0).
      compute sub999=csum(sub999).
      do if (sub999=0).
        loop i = 1 to items. 
          compute pre = {make(nrow(y),1,1), y(:,i)}.
          compute b = inv(t(pre)*pre)*t(pre)*y(:,(items+1)).
          compute res(i,:)={b(1,1),(b(1,1)+b(2,1)),b(2,1)}.
        end loop.
        print res/title = "Mean reliability of all possible subset scales with and without item:"
        /clabels = "Without", "With", "Gain"/rnames = vnames/format = !decimals.        
      end if.
      do if (save=1).
        compute vnames = {vnames, "reliab", "r"}.
        save y/outfile = */names = vnames.
        compute savedit=1.
      end if.  
    end if.
  end if.
end if.

print/title="********************************************************************************".
do if (toofew=1).
  print/title="ERROR: Too few items for the analysis you have requested.".
end if.
/* do if (toomany=1) */.
/*  print/title="ERROR: Too many items. 15 is the maximum allowed." */.
/* end if */.
do if (criterr=0).
  print n/title = "Number of cases:".
  do if (m999=1).
    print/title="NOTE: -99 indicates that the algorithm for calculating this quantity generated an".
    print/title="error. This is probably due to including an item in the scale that is very "/space=0.
    print/title="weakly or negatively correlated with the others."/space=0.
  end if.
  do if (sub999>0).
    print/title="NOTE: Because reliability could not be estimated for some subset scales, the".
    print/title="table of mean reliability with and without item is not generated."/space=0.
  end if.
  do if (bootok=1).
    print conf/title "Percentage of confidence for bootstrap confidence interval:".
    do if (boots >=1000 and !boot >= 1000).
      print boots/title="NOTE: The number of bootstrap samples for confidence interval:".
    end if.
    do if (!boot < 1000).
      print/title="NOTE: 1000 bootstrap samples is the minimum, so 1000 was used for".
      print/title="construction of the confidence interval."/space=0.
    end if.
    do if (conferr=1).
      print/title="NOTE: Only 90%, 95%, or 99% confidence intervals are available. 95% was used.".
    end if.
    do if (bdboot > 0).
      print/title="NOTE: Some bootstrap samples had to be replaced due to computational".
      print/title="problems. Interpret confidence interval with caution. The number of"/space=0.
      print bdboot/title="bootstrap samples replaced was:"/space=0.
    end if.
    do if ((!quote(!seed) <> "random")).
      print !seed/title="Seed used for the random number generator:"/format=F8.0.
    end if.
  end if.
  do if (complete=0 and boots > 0).
    print/title="NOTE: Bootstrapping could not be completed due to computational problems so".
    print/title="no bootstrap confidence interval is provided."/space=0.
  end if.
  do if (savedit=1).
    print/title="NOTE: A file of all subsets was generated at your request.".
  end if.
end if.
END MATRIX.
!ifend
!ENDDEFINE.


